--
--    FILE:     Tectonics.lua
--    AUTHOR:  Laurent Di Cesare
--    PURPOSE: Map script simulating plate tectonics for nice mountain ranges
------------------------------------------------------------------------------
--    Copyright (c) 2010 Laurent Di Cesare All rights reserved.
------------------------------------------------------------------------------
--      HISTORY
--      VERSION: 6
--               Added starting plots override to force at least two players per continent, except if 90% water is chosen.
--               Also added standard resources variable in the process.
--      VERSION: 5
--               Supports multiplayer.
--               Added sea stripes near the poles so it's less likely to have continents span from pole to pole
--               Changed number of plates to get more continents on 60/70% water settings
--      VERSION: 4
--               Added map option for water proportion.
--      VERSION: 3
--               Corrected a  bug where map generation would fail because some heights had no plot associated (use (plotNumbers[i] or 0) in ComputeHeight).
--      VERSION: 2
--               Corrected the require away. Added not advanced map.
--      VERSION: 1
------------------------------------------------------------------------------

include("math")
include("MapGenerator")
include("FeatureGenerator")
include("TerrainGenerator")

function GetMapScriptInfo()
    local world_age, temperature, rainfall, sea_level, resources = GetCoreMapOptions()
    return {
        Name = "Tectonics",
        Description = "Tectonics",
        IconIndex = 5,
        IsAdvancedMap = false,
        SupportsMultiplayer = true,
        CustomOptions = {
            {
                Name = "Water proportion",
                Values = {
                    "Arid (10%)",
                    "Little water (30%)",
                    "Half (50%)",
                    "Many seas (60%)",
                    "Earthlike (70%)",
                    "Mostly seas (90%)",
                },
                DefaultValue = 4,
                SortPriority = 1,
            },
            resources,
        },
    }
end
local Wait = function(arg) -- debbugging function to be able to use the *** console
    if arg then print(arg) end
    --[[
    local silly = 0
    while silly < 9900000 do
        silly = silly + 1
    end
    --]]
end
local Voronoi =
{
    plotTypes = {}, -- map generation result
    plate = {}, -- plate of plot
    plateCollision = {}, -- base height increase of each plate if it collides with someone
    heightMap = {},
    collisionHeight = 40, -- height added when 2 plates collide
    seaLevel = 0, -- base sea altitude
    landLevel = 20, -- base land altitude
    seaNoise = 6, -- max random sea altitude
    landNoise = 9, -- max random land altitude
    landPlotHeight = 20, -- default land height, shouldn't be used if proportions are used instead
    hillPlotHeight = 30,
    mountainPlotHeight = 44,
    seaProportion = 0.5,
    landProportion = 0.3,
    hillProportion = 0.15,
    seaOption =  {0.1,0.3,0.5,0.6,0.7,0.9},
    landOption = {0.6,0.5,0.35,0.3,0.2,0.07},
    hillOption = {0.25,0.15,0.1,0.05,0.05,0.02},
    Create = function(self)
        self:InitWaterProportion()
        self.iW, self.iH = Map.GetGridSize()
        self:GenerateSeeds()
        self:FillPlates()
        self:FillPlateAltitudes()
        self:MovePlates()
        self:Blur()
        self:MovePlates()
        self:ComputePlots()
        return self.plotTypes
    end,
    InitWaterProportion = function(self)
        local option = Map.GetCustomOption(1)
        self.seaProportion = self.seaOption[option]
        self.landProportion = self.landOption[option]
        self.hillProportion = self.hillOption[option]
    end,
    AvoidLandAtPoles = function(self)
        local lastRow = (self.iH - 1)*self.iW
        for plotIndex = 1, self.iW do
            self.plate[plotIndex] = 0
            self.plate[lastRow + plotIndex] = 0
        end
    end,
    GenerateSeeds = function(self)
        self.plateNumber = self.iH / 20 + self.iW / 10 + (self.iH * self.iW) / 100
        self.seaPlates = math.floor( self.plateNumber * self.seaProportion ) + 1
        self.expectedSea = math.floor( self.iH * self.iW * self.seaProportion )
        self.expectedLand = math.floor( self.iH * self.iW * self.landProportion )
        self.expectedHill = math.floor( self.iH * self.iW * self.hillProportion )
        self:AvoidLandAtPoles()
        for i = 1, self.seaPlates - 1 do
            local plotIndex = Map.Rand(self.iH, "y for seed") * self.iW + Map.Rand(self.iW, "x for seed")
            while self.plate[plotIndex] do
                plotIndex = Map.Rand(self.iH, "y for seed") * self.iW + Map.Rand(self.iW, "x for seed")
            end
            self.plate[plotIndex] = i
        end
        local landHemisphere = Map.Rand(10, "hemisphere for more land")
        local yLandOffset = -1
        if landHemisphere > 4 then yLandOffset = 1 end
        for i = self.seaPlates, self.plateNumber do
            local hemisphere = Map.Rand(100, "hemisphere for land plate")
            if hemisphere > 75 then yLandOffset = yLandOffset * -1 end
            local yMin = math.floor(self.iH/10)
            local yVar = math.floor(self.iH/2)
            local RandY = function(self)
                local y = yMin + Map.Rand(yVar, "y for seed")
                if yLandOffset > 0 then
                    y = self.iH - y
                end
                return y
            end
            local plotIndex = RandY(self) * self.iW + Map.Rand(self.iW, "x for seed")
            while self.plate[plotIndex] do
                plotIndex = RandY(self) * self.iW + Map.Rand(self.iW, "x for seed")
            end
            self.plate[plotIndex] = i
        end
    end,
    South = function(self, x, y)
        if y == 0 then return nil end
        return (y-1)*self.iW + x + 1
    end,
    North = function(self, x, y)
        if y == self.iH - 1 then return nil end
        return (y+1)*self.iW + x + 1
    end,
    West = function(self, x, y)
        if x == 0 then return y*self.iW + self.iW - 1 +1 end
        return y*self.iW + x - 1 + 1
    end,
    East = function(self, x, y)
        if x == self.iW - 1 then return y*self.iW + 1 end
        return y*self.iW + x + 1 + 1
    end,
    NorthWest = function(self, x, y)
        if y == self.iH - 1 then return nil end
        return self:West(x, y + 1)
    end,
    SouthWest = function(self, x, y)
        if y == 0 then return nil end
        return self:West(x, y - 1)
    end,
    Neighbours = function(self, x, y)
        local result = {}
        result["South"] = self:South(x, y)
        result["North"] = self:North(x, y)
        result["West"] = self:West(x, y)
        result["East"] = self:East(x, y)
        result["NorthWest"] = self:NorthWest(x, y)
        result["SouthWest"] = self:SouthWest(x, y)
        return result
    end,
    FillPlates = function( self )
        local continue = true
        while continue do
            continue = false
            local newPlate = {}
            for y = 0, self.iH - 1 do
                for x = 0, self.iW - 1 do
                    local plotIndex = y * self.iW + x + 1
                    local currentPlate = self.plate[plotIndex]
                    -- testing currentPlate first would work and be slightly faster I believe
                    -- but it'd look harder to understand what I'm doing
                    newPlate[plotIndex] = newPlate[plotIndex] or currentPlate
                    if currentPlate then
                        for _, i in pairs(self:Neighbours(x, y)) do
                            if not self.plate[i] then
                                if Map.Rand(10, "randomness in plate size") > 4 then
                                    newPlate[i] = currentPlate
                                end
                            end
                        end
                    end
                end
            end
            self.plate = newPlate
            for i = 1, self.iH * self.iW do
                if not newPlate[i] then continue = true break end
            end
        end
    end,
    FillPlateAltitudes = function(self)
        for i = 0, self.seaPlates - 1 do
            self.plateCollision[i] = self.seaLevel + Map.Rand(self.seaNoise, "sea level collision")
        end
        for i = self.seaPlates, self.plateNumber do
            self.plateCollision[i] = self.collisionHeight
        end
        for plotIndex = 1, self.iH * self.iW do
            if self.plate[plotIndex] < self.seaPlates then
                self.heightMap[plotIndex] = self.seaLevel + Map.Rand(self.seaNoise, "sea level plate variation")
            else
                self.heightMap[plotIndex] = self.landLevel + Map.Rand(self.landNoise, "land level plate variation")
            end
        end
    end,
    MovePlates = function(self)
        local direction = {}
        for i = 0, self.plateNumber do
            direction[i] = 1 + Map.Rand(5, "direction for plate")
        end
        local directionFunctions = { self.North, self.South, self.East, self.West, self.NorthWest, self.SouthWest }
        for y = 0, self.iH - 1 do
            for x = 0, self.iW - 1 do
                local plotIndex = y * self.iW + x + 1
                local currentPlate = self.plate[plotIndex]
                local dirFunction = directionFunctions[direction[currentPlate]]
                local destinationPlot = dirFunction(self, x, y)
                if destinationPlot and (self.plate[destinationPlot] ~= currentPlate) then
                    self.heightMap[plotIndex] = self.heightMap[plotIndex] + self.plateCollision[currentPlate] + self.plateCollision[self.plate[destinationPlot]] + Map.Rand(self.seaNoise, "collision noise")
                end
            end
        end
    end,
    Blur = function(self)
        local heightMap = {}
        for y = 0, self.iH - 1 do
            for x = 0, self.iW - 1 do
                local plotIndex = y * self.iW + x + 1
                local height = self.heightMap[plotIndex]
                local neighbours = 1
                for dir, i in pairs(self:Neighbours(x, y)) do
                    height = height + self.heightMap[i]
                    neighbours = neighbours + 1
                end
                heightMap[plotIndex] = math.floor(height/neighbours)
            end
        end
        self.heightMap  = heightMap
    end,
    ComputeHeights = function(self)
        local plotNumbers = {}
        local totalPlots = self.iH * self.iW
        local min = math.huge
        local max = 0
        for plotIndex = 1, totalPlots do
            local height = self.heightMap[plotIndex]
            if not plotNumbers[height] then plotNumbers[height] = 1 end
            plotNumbers[height] = plotNumbers[height] + 1
            max = math.max(max, height)
            min = math.min(min, height)
        end
        local sea = 0
        local land = 0
        local hill = 0
        for i = min, max do
            sea = sea + (plotNumbers[i] or 0)
            if sea > self.expectedSea then self.landPlotHeight = i break end
        end
        self.hillPlotHeight = self.landPlotHeight + 1
        for i = self.landPlotHeight + 1, max do
            land = land + (plotNumbers[i] or 0)
            if land > self.expectedLand then self.hillPlotHeight = i break end
        end
        self.mountainPlotHeight = max
        for i = self.hillPlotHeight + 1, max do
            hill = hill + (plotNumbers[i] or 0)
            if hill > self.expectedHill then self.mountainPlotHeight = i break end
        end
    end,
    CleanIsolatedLakes = function(self)
        for y = 0, self.iH - 1 do
            for x = 0, self.iW - 1 do
                local plotIndex = y * self.iW + x + 1
                if self.plotTypes[plotIndex] == PlotTypes.PLOT_OCEAN then
                    local water = false
                    for _, index in pairs(self:Neighbours(x, y)) do
                        if self.plotTypes[index] == PlotTypes.PLOT_OCEAN then
                            water = true
                            break
                        end
                    end
                    if not water then
                        self.plotTypes[plotIndex] = PlotTypes.PLOT_HILL
                    end
                end
            end
        end
    end,
    ComputePlots = function(self)
        self:ComputeHeights()
        for y = 0, self.iH - 1 do
            for x = 0, self.iW - 1 do
                local plotIndex = y * self.iW + x + 1
                local height = self.heightMap[plotIndex]
                if height < self.landPlotHeight then
                    self.plotTypes[plotIndex] = PlotTypes.PLOT_OCEAN
                elseif height < self.hillPlotHeight then 
                    self.plotTypes[plotIndex] = PlotTypes.PLOT_LAND
                elseif height < self.mountainPlotHeight then
                    self.plotTypes[plotIndex] = PlotTypes.PLOT_HILLS
                else
                    if Map.Rand(100, "mountain pass" ) > 90 then
                        self.plotTypes[plotIndex] = PlotTypes.PLOT_HILLS
                    else
                        self.plotTypes[plotIndex] = PlotTypes.PLOT_MOUNTAIN
                    end
                end
            end
        end
        self:CleanIsolatedLakes()
    end,
}
----------------------------------------------------------------------------------
function GeneratePlotTypes()
    print("Setting Plot Types (Tectonics) ...")

    local plotTypes = Voronoi:Create()
    SetPlotTypes(plotTypes)
    local args = { bExpandCoasts = false }
    GenerateCoasts(args) -- will have to look into that as I want to avoid coasts spanning into oceans as they currently do too easily

end
----------------------------------------------------------------------------------
function GenerateTerrain()
    print("Generating Terrain (Using default for the moment) ...")
    
    local terraingen = TerrainGenerator.Create()
    local terrainTypes = terraingen:GenerateTerrain()
        
    SetTerrainTypes(terrainTypes)
end
----------------------------------------------------------------------------------
function AddFeatures()
    print("Adding Features (using default implementation) ...")
    
    local featuregen = FeatureGenerator.Create()
    featuregen:AddFeatures()
end
----------------------------------------------------------------------------------
local TectonicsGenerateRegions = function(self, args)
    -- This function is more or less a copy of the default generation method except
    -- it forces players to be at least 2 per landmass.
    -- My changes are mostly everything that doesn't end with a ";"
    -- Comments have been removed to make it easier to read and find out error syntaxes...
    local args = args or {}
    local iW, iH = Map.GetGridSize()
    self.resource_setting = args.resources or 2

    self.iNumCivs, self.iNumCityStates, self.player_ID_list, self.bTeamGame, self.teams_with_major_civs, self.number_civs_per_team = GetPlayerAndTeamInfo();
    self.iNumCityStatesUnassigned = self.iNumCityStates;

    local best_areas = {};
    local globalFertilityOfLands = {};

    local iGlobalFertilityOfLands = 0;
    local iNumLandPlots = 0;
    local iNumLandAreas = 0;
    local land_area_IDs = {};
    local land_area_plots = {};
    local land_area_fert = {};
    for x = 0, iW - 1 do
        for y = 0, iH - 1 do
            local i = y * iW + x + 1;
            local plot = Map.GetPlot(x, y);
            if not plot:IsWater() then 
                iNumLandPlots = iNumLandPlots + 1;
                local iArea = plot:GetArea();
                local plotFertility = self:MeasureStartPlacementFertilityOfPlot(x, y, true);
                iGlobalFertilityOfLands = iGlobalFertilityOfLands + plotFertility;
                --
                if TestMembership(land_area_IDs, iArea) == false then
                    iNumLandAreas = iNumLandAreas + 1;
                    table.insert(land_area_IDs, iArea);
                    land_area_plots[iArea] = 1;
                    land_area_fert[iArea] = plotFertility;
                else
                    land_area_plots[iArea] = land_area_plots[iArea] + 1;
                    land_area_fert[iArea] = land_area_fert[iArea] + plotFertility;
                end
            end
        end
    end

    local interim_table = {};
    for loop_index, data_entry in pairs(land_area_fert) do
        table.insert(interim_table, data_entry);
    end
    table.sort(interim_table);
    local iNumRelevantLandAreas = math.min(iNumLandAreas, self.iNumCivs);
    local best_areas = {};
    local end_of_interim_table = table.maxn(interim_table);
    for areaTestLoop = end_of_interim_table, (end_of_interim_table - iNumRelevantLandAreas + 1), -1 do
        for loop_index, AreaID in ipairs(land_area_IDs) do
            if interim_table[areaTestLoop] == land_area_fert[land_area_IDs[loop_index]] then
                table.insert(best_areas, AreaID);
                table.remove(land_area_IDs, landLoop);
                break
            end
        end
    end
    local inhabitedAreaIDs = {};
    local numberOfCivsPerArea = table.fill(0, iNumRelevantLandAreas);
    local odd = true
    local loopMax = math.floor(self.iNumCivs / 2)
    if (self.iNumCivs / 2) == loopMax then
        odd = false
    end
    for civToAssign = 1, loopMax do
        local bestRemainingArea;
        local bestRemainingFertility = 0;
        local bestAreaTableIndex;
        for area_loop, AreaID in ipairs(best_areas) do
            local thisLandmassCurrentFertility = land_area_fert[AreaID] / (1 + numberOfCivsPerArea[area_loop]);
            if thisLandmassCurrentFertility > bestRemainingFertility then
                bestRemainingArea = AreaID;
                bestRemainingFertility = thisLandmassCurrentFertility;
                bestAreaTableIndex = area_loop;
            end
        end
        local addedPlayers = 2
        if odd then
            addedPlayers = 3
            odd = false
        end
        numberOfCivsPerArea[bestAreaTableIndex] = numberOfCivsPerArea[bestAreaTableIndex] + addedPlayers;
        if TestMembership(inhabitedAreaIDs, bestRemainingArea) == false then
            table.insert(inhabitedAreaIDs, bestRemainingArea);
        end
    end
    local iNumInhabitedLandmasses = table.maxn(inhabitedAreaIDs);
    for loop, currentLandmassID in ipairs(inhabitedAreaIDs) do
        local landmass_data = ObtainLandmassBoundaries(currentLandmassID);
        local iWestX = landmass_data[1];
        local iSouthY = landmass_data[2];
        local iEastX = landmass_data[3];
        local iNorthY = landmass_data[4];
        local iWidth = landmass_data[5];
        local iHeight = landmass_data[6];
        local wrapsX = landmass_data[7];
        local wrapsY = landmass_data[8];
        local fert_table, fertCount, plotCount = self:MeasureStartPlacementFertilityOfLandmass(currentLandmassID, 
        iWestX, iEastX, iSouthY, iNorthY, wrapsX, wrapsY);
        local rect_table = {iWestX, iSouthY, iWidth, iHeight, currentLandmassID, fertCount, plotCount};
        iNumCivsOnThisLandmass = numberOfCivsPerArea[loop];
        if iNumCivsOnThisLandmass > 0 and iNumCivsOnThisLandmass <= 22 then -- valid number of civs.
            self:DivideIntoRegions(iNumCivsOnThisLandmass, fert_table, rect_table)
        else
            print("Invalid number of civs assigned to a landmass: ", iNumCivsOnThisLandmass);
        end
    end
    self:CustomOverride() -- not that it matters but in case someone overrides tectonics...
end
----------------------------------------------------------------------------------
function StartPlotSystem()
    -- Get Resources setting input by user.
    local res = Map.GetCustomOption(2)
    if res == 6 then
        res = 1 + Map.Rand(3, "Random Resources Option - Lua")
    end

    print("Creating start plot database (MapGenerator.Lua)")
    local start_plot_database = AssignStartingPlots.Create()

    if (Map.GetCustomOption(1) ~= 6) then
        -- keep default algorithm only for 90% water,
        -- otherwise force players to be at least 2 per landmass
        -- otherwise even a moron cna win on deity if alone on his continent
        -- which is not fun and I don't want to ever be alone on my continent.
        start_plot_database.GenerateRegions = TectonicsGenerateRegions
    end
    
    start_plot_database:GenerateRegions({resources = res})

    print("Choosing start locations for civilizations (Tectonics)");
    start_plot_database:ChooseLocations()
    
    print("Normalizing start locations and assigning them to Players (MapGenerator.Lua)");
    start_plot_database:BalanceAndAssign()

    print("Placing Natural Wonders (MapGenerator.Lua)");
    start_plot_database:PlaceNaturalWonders()

    print("Placing Resources and City States (MapGenerator.Lua)");
    start_plot_database:PlaceResourcesAndCityStates()
end
